@@ -9,6 +9,8 @@ module Data.Conduit.Network.TLS
       TLSConfig
     , tlsConfigBS
     , tlsConfig
+    , tlsConfigChainBS
+    , tlsConfigChain
     , tlsHost
     , tlsPort
 --    , tlsCertificate
@@ -16,6 +18,7 @@ module Data.Conduit.Network.TLS
     , tlsNeedLocalAddr
     , tlsAppData
     , runTCPServerTLS
+    , runGeneralTCPServerTLS
     , runTCPServerStartTLS
       -- * Client
     , TLSClientConfig
@@ -32,7 +35,7 @@ module Data.Conduit.Network.TLS
 
 import Prelude hiding (FilePath, readFile)
 import Control.Applicative ((<$>), (<*>))
-import Control.Monad (forever)
+import Control.Monad (forever, void)
 import Filesystem.Path.CurrentOS (FilePath)
 import Filesystem (readFile)
 import qualified Data.ByteString.Lazy as L
@@ -58,19 +61,24 @@ import Control.Monad.Trans.Control
 import Data.Default
 
 
-makeCertDataPath :: FilePath -> FilePath -> TlsCertData
-makeCertDataPath certPath keyPath = TlsCertData (readFile certPath) (readFile keyPath)
-
-makeCertDataBS :: S.ByteString -> S.ByteString -> TlsCertData
-makeCertDataBS certBS keyBS = TlsCertData (return certBS) (return keyBS)
+makeCertDataPath :: FilePath -> [FilePath] -> FilePath -> TlsCertData
+makeCertDataPath certPath chainCertPaths keyPath =
+    TlsCertData
+      (readFile certPath)
+      (mapM readFile chainCertPaths)
+      (readFile keyPath)
 
+makeCertDataBS :: S.ByteString -> [S.ByteString] -> S.ByteString ->
+                  TlsCertData
+makeCertDataBS certBS chainCertsBS keyBS =
+    TlsCertData (return certBS) (return chainCertsBS) (return keyBS)
 
 tlsConfig :: HostPreference
           -> Int -- ^ port
           -> FilePath -- ^ certificate
           -> FilePath -- ^ key
           -> TLSConfig
-tlsConfig a b c d = TLSConfig a b (makeCertDataPath c d) False
+tlsConfig a b c d = tlsConfigChain a b c [] d
 
 
 -- | allow to build a server config directly from raw bytestring data (exact same
@@ -81,8 +89,30 @@ tlsConfigBS :: HostPreference
             -> S.ByteString -- ^ Certificate raw data
             -> S.ByteString -- ^ Key file raw data
             -> TLSConfig
-tlsConfigBS a b c d = TLSConfig a b (makeCertDataBS c d ) False
+tlsConfigBS a b c d = tlsConfigChainBS a b c [] d
 
+-- | Like 'tlsConfig', but also allow specifying chain certificates.
+--
+-- Since 1.1.1
+tlsConfigChain :: HostPreference
+               -> Int -- ^ Port
+               -> FilePath -- ^ Certificate
+               -> [FilePath] -- ^ Chain certificates
+               -> FilePath -- ^ Key
+               -> TLSConfig
+tlsConfigChain a b c d e = TLSConfig a b (makeCertDataPath c d e) False
+
+
+-- | Like 'tlsConfigBS', but also allow specifying chain certificates.
+--
+-- Since 1.1.1
+tlsConfigChainBS :: HostPreference
+                 -> Int          -- ^ Port
+                 -> S.ByteString -- ^ Certificate raw data
+                 -> [S.ByteString] -- ^ Chain certificate raw data
+                 -> S.ByteString -- ^ Key file raw data
+                 -> TLSConfig
+tlsConfigChainBS a b c d e = TLSConfig a b (makeCertDataBS c d e) False
 
 serverHandshake :: Socket -> TLS.Credentials -> IO (TLS.Context)
 serverHandshake socket creds = do
@@ -132,6 +162,18 @@ runTCPServerTLS TLSConfig{..} app = do
 
 type ApplicationStartTLS = (AppData, (AppData -> IO ()) -> IO ()) -> IO ()
 
+-- | Like 'runTCPServerTLS', but monad can be any instance of 'MonadBaseControl' 'IO'.
+--
+-- Note that any changes to the monadic state performed by individual
+-- client handlers will be discarded. If you have mutable state you want
+-- to share among multiple handlers, you need to use some kind of mutable
+-- variables.
+--
+-- Since 1.1.2
+runGeneralTCPServerTLS :: MonadBaseControl IO m => TLSConfig -> (AppData -> m ()) -> m ()
+runGeneralTCPServerTLS config app = liftBaseWith $ \run ->
+  runTCPServerTLS config $ void . run . app
+
 -- | run a server un-crypted but also pass a call-back to trigger a StartTLS handshake
 -- on the underlying connection
 --
@@ -162,6 +204,9 @@ runTCPServerStartTLS TLSConfig{..} app = do
                   , appLocalAddr' = mlocal
 #if MIN_VERSION_streaming_commons(0,1,6)
                   , appCloseConnection' = sClose socket
+#endif
+#if MIN_VERSION_streaming_commons(0,1,12)
+                  , appRawSocket' = Just socket
 #endif
                   }
                 -- wrap up the current connection with TLS
@@ -197,6 +242,9 @@ tlsAppData ctx addr mlocal = AppData
     , appLocalAddr' = mlocal
 #if MIN_VERSION_streaming_commons(0,1,6)
     , appCloseConnection' = TLS.contextClose ctx
+#endif
+#if MIN_VERSION_streaming_commons(0,1,12)
+    , appRawSocket' = Nothing
 #endif
     }
 
@@ -210,8 +258,8 @@ ciphers =
     ]
 
 readCreds :: TlsCertData -> IO TLS.Credentials
-readCreds (TlsCertData iocert iokey) =
-    (TLS.credentialLoadX509FromMemory <$> iocert <*> iokey)
+readCreds (TlsCertData iocert iochains iokey) =
+    (TLS.credentialLoadX509ChainFromMemory <$> iocert <*> iochains <*> iokey)
     >>= either
         (error . ("Error reading TLS credentials: " ++))
         (return . TLS.Credentials . return)
@@ -305,6 +353,9 @@ runTLSClient TLSClientConfig {..} app = do
             , appLocalAddr' = Nothing
 #if MIN_VERSION_streaming_commons(0,1,6)
             , appCloseConnection' = NC.connectionClose conn
+#endif
+#if MIN_VERSION_streaming_commons(0,1,12)
+            , appRawSocket' = Nothing
 #endif
             })
 
@@ -335,6 +386,9 @@ runTLSClientStartTLS TLSClientConfig {..} app = do
             , appLocalAddr' = Nothing
 #if MIN_VERSION_streaming_commons(0,1,6)
             , appCloseConnection' = NC.connectionClose conn
+#endif
+#if MIN_VERSION_streaming_commons(0,1,12)
+            , appRawSocket' = Nothing
 #endif
             }
             , \app' -> do
@@ -346,6 +400,9 @@ runTLSClientStartTLS TLSClientConfig {..} app = do
                    , appLocalAddr' = Nothing
 #if MIN_VERSION_streaming_commons(0,1,6)
                    , appCloseConnection' = NC.connectionClose conn
+#endif
+#if MIN_VERSION_streaming_commons(0,1,12)
+                    , appRawSocket' = Nothing
 #endif
                    }
             )
